#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Curtis Yu
@contact: cuyu@splunk.com
@since: 13/12/2017
"""
import logging
import sys
from colorama import Fore, Back, Style

# Python 2+3 compatibility settings for logger
bytes_type = bytes
if sys.version_info >= (3,):
    unicode_type = str
    basestring_type = str
    xrange = range
else:
    # The names unicode and basestring don't exist in py3 so silence flake8.
    unicode_type = unicode  # noqa
    basestring_type = basestring  # noqa


def _safe_unicode(s):
    try:
        if isinstance(s, (unicode_type, type(None))):
            return s
        if not isinstance(s, bytes):
            raise TypeError(
                "Expected bytes, unicode, or None; got %r" % type(s))
        return s.decode("utf-8")
    except UnicodeDecodeError:
        return repr(s)


class LogFormatter(logging.Formatter):
    """
    Log formatter used in Tornado. Key features of this formatter are:
    * Color support when logging to a terminal that supports it.
    * Timestamps on every log line.
    * Robust against str/bytes encoding problems.
    """
    DEFAULT_FORMAT = '%(color)s[%(asctime)s] %(levelname)1.1s - %(name)s%(empty)s| %(end_color)s %(message)s'
    DEFAULT_DATE_FORMAT = '%y-%m-%d %H:%M:%S'
    DEFAULT_STYLES = {
        logging.DEBUG: Style.NORMAL,
        logging.INFO: Style.NORMAL,
        logging.WARNING: Style.BRIGHT,
        logging.ERROR: Style.BRIGHT,
        logging.CRITICAL: Style.BRIGHT,
    }

    def __init__(self,
                 color,
                 fmt=DEFAULT_FORMAT,
                 datefmt=DEFAULT_DATE_FORMAT):
        r"""
        :arg bool color: Enables color support.
        :arg string fmt: Log message format.
          It will be applied to the attributes dict of log records. The
          text between ``%(color)s`` and ``%(end_color)s`` will be colored
          depending on the level if color support is on.
        :arg dict colors: color mappings from logging level to terminal color
          code
        :arg string datefmt: Datetime format.
          Used for formatting ``(asctime)`` placeholder in ``prefix_fmt``.
        .. versionchanged:: 3.2
           Added ``fmt`` and ``datefmt`` arguments.
        """
        logging.Formatter.__init__(self, datefmt=datefmt)
        self._color = color
        self._fmt = fmt
        self._normal = Style.RESET_ALL

    def format(self, record):
        try:
            message = record.getMessage()
            assert isinstance(message,
                              basestring_type)  # guaranteed by logging
            # Encoding notes:  The logging module prefers to work with character
            # strings, but only enforces that log messages are instances of
            # basestring.  In python 2, non-ascii bytestrings will make
            # their way through the logging framework until they blow up with
            # an unhelpful decoding error (with this formatter it happens
            # when we attach the prefix, but there are other opportunities for
            # exceptions further along in the framework).
            #
            # If a byte string makes it this far, convert it to unicode to
            # ensure it will make it out to the logs.  Use repr() as a fallback
            # to ensure that all byte strings can be converted successfully,
            # but don't do it by default so we don't add extra quotes to ascii
            # bytestrings.  This is a bit of a hacky place to do this, but
            # it's worth it since the encoding errors that would otherwise
            # result are so useless (and tornado is fond of using utf8-encoded
            # byte strings wherever possible).
            record.message = _safe_unicode(message)
        except Exception as e:
            record.message = "Bad message (%r): %r" % (e, record.__dict__)

        record.asctime = self.formatTime(record, self.datefmt)

        if record.levelno in self.DEFAULT_STYLES:
            record.color = self.DEFAULT_STYLES[record.levelno] + self._color
            record.end_color = self._normal
        else:
            record.color = record.end_color = ''

        space_number = 12 - len(record.name)
        record.empty = '' if space_number <= 0 else ' ' * space_number

        formatted = self._fmt % record.__dict__

        if record.exc_info:
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            # exc_text contains multiple lines.  We need to _safe_unicode
            # each line separately so that non-utf8 bytes don't cause
            # all the newlines to turn into '\n'.
            lines = [formatted.rstrip()]
            lines.extend(
                _safe_unicode(ln) for ln in record.exc_text.split('\n'))
            formatted = '\n'.join(lines)
        return formatted.replace("\n", "\n    ")


def setup_logger(name=None, color=None, level=logging.DEBUG):
    _logger = logging.getLogger(name or __name__)
    _logger.propagate = False
    _logger.setLevel(level)

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(level)
    stream_handler.setFormatter(LogFormatter(color))
    _logger.addHandler(stream_handler)

    return _logger


if __name__ == '__main__':
    from colorama import init
    init()
    logger = setup_logger('TestLogger', color=Fore.BLUE)
    logger.info('hello')
    logger.error('errrrr')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Curtis Yu
@contact: cuyu@splunk.com
@since: 06/12/2017
"""
import stat
import docker
import os
import paramiko
from common import tar_cz, tar_xz
from logger import setup_logger
from colorama import Fore

_LOGGER = setup_logger('Client', color=Fore.CYAN)


class Client(object):
    """
    A client to wrap both ssh and docker client
    """

    def __init__(self, host, ssh_username=None, ssh_password=None, ssh_port=22):
        if ssh_username and ssh_password:
            self._ssh_client = paramiko.SSHClient()
            self._ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            self._ssh_client.connect(host, ssh_port, ssh_username, ssh_password)
            self._sftp = paramiko.SFTPClient.from_transport(self._ssh_client.get_transport())
            self._docker_client = None
        else:
            self._ssh_client = None
            self._docker_client = docker.from_env()
            self._docker_container = self._docker_client.containers.get(host)

    def execute(self, command):
        _LOGGER.info('executing command: {0}'.format(command))
        if self._ssh_client:
            # Return (stdin, stdout, stderr) which are file-like objects
            stdin, stdout, stderr = self._ssh_client.exec_command(command)
            stdout = ''.join(stdout.readlines())
            _LOGGER.info('stdout: \n{0}'.format(stdout))
            return stdin, stdout, stderr
        else:
            # Return generator or str
            result = self._docker_container.exec_run(command)
            if isinstance(result, str):
                stdout = result
            else:
                stdout = ''
                for line in result:
                    stdout += line + '\n'
            _LOGGER.info('stdout: \n{0}'.format(stdout))
            return result

    def send_files(self, local_path, remote_path):
        _LOGGER.info('sending files from {0} (local) to {1} (remote)'.format(local_path, remote_path))
        if self._ssh_client:
            if os.path.isdir(local_path):
                self._ssh_send_folder(local_path, remote_path)
            else:
                self._sftp.put(local_path, remote_path)
        else:
            data = tar_cz(local_path)
            self._docker_container.put_archive(os.path.dirname(remote_path), data)

    def _ssh_send_folder(self, local_path, remote_path):
        self._ssh_mkdir(remote_path, ignore_existing=True)

        for item in os.listdir(local_path):
            if os.path.isfile(os.path.join(local_path, item)):
                self._sftp.put(os.path.join(local_path, item), '%s/%s' % (remote_path, item))
            else:
                self._ssh_mkdir('%s/%s' % (remote_path, item), ignore_existing=True)
                self._ssh_send_folder(os.path.join(local_path, item), '%s/%s' % (remote_path, item))

    def _ssh_mkdir(self, path, mode=511, ignore_existing=False):
        ''' Augments mkdir by adding an option to not fail if the folder exists  '''
        try:
            self._sftp.mkdir(path, mode)
        except IOError:
            if ignore_existing:
                pass
            else:
                raise

    def fetch_files(self, remote_path, local_path):
        """
        :param remote_path: e.g. '/tmp/project/', '/tmp/debug.log'
        :param local_path: e.g. '/tmp/project/', '/tmp/debug.log'
        """
        _LOGGER.info('fetching files from {0} (remote) to {1} (local)'.format(remote_path, local_path))
        local_dir = os.path.dirname(local_path)
        assert os.path.exists(local_dir)

        if self._ssh_client:
            if stat.S_ISDIR(self._sftp.stat(remote_path).st_mode):
                self._ssh_fetch_folder(remote_path, os.path.join(local_dir, os.path.basename(remote_path)))
            else:
                self._sftp.get(remote_path, local_path)
        else:
            response, _ = self._docker_container.get_archive(remote_path)
            tar_xz(response.data, local_dir)

    def _ssh_fetch_folder(self, remote_path, local_path):
        if not os.path.exists(local_path):
            os.mkdir(local_path)

        for f in self._sftp.listdir(remote_path):
            if stat.S_ISDIR(self._sftp.stat(os.path.join(remote_path, f)).st_mode):
                self._ssh_fetch_folder(os.path.join(remote_path, f), os.path.join(local_path, f))
            else:
                self._sftp.get(os.path.join(remote_path, f), os.path.join(local_path, f))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Curtis Yu
@contact: cuyu@splunk.com
@since: 06/12/2017
"""
import socket
import tarfile
import os
import psutil

try:
    from cStringIO import StringIO as BIO
except ImportError:  # python 3
    from io import BytesIO as BIO


def get_host_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    finally:
        s.close()

    return ip


def get_unoccupied_port():
    s = socket.socket()
    s.bind(('', 0))

    return s.getsockname()[1]


def get_pid_by_name(name):
    pid = []
    for p in psutil.process_iter():
        if hasattr(p, 'name') and p.name() == name:
            pid.append(p.pid)
    return pid


def is_port_in_use(pid, port):
    p = psutil.Process(pid)
    for conn in p.connections():
        if conn.laddr[1] == port:
            return True
    return False


def tar_cz(*path):
    """tar_cz(*path) -> bytes
    Compress a sequence of files or directories in memory.
    The resulting string could be stored as a .tgz file."""
    file_out = BIO()
    tar = tarfile.open(mode="w:gz", fileobj=file_out)
    for p in path:
        tar.add(p, arcname=os.path.basename(p))
    tar.close()
    return file_out.getvalue()


def tar_xz(stringz, folder="."):
    """tar_xz(stringz, folder = ".") -> None
    Uncompress a string created by tar_cz in a given directory."""
    file_in = BIO(stringz)
    tar = tarfile.open(mode="r", fileobj=file_in)
    tar.extractall(folder)



import os
import xml.etree.ElementTree as et
import argparse
import pkgutil
import shutil
from subprocess import call
import time
from common import get_host_ip, get_unoccupied_port, is_port_in_use, get_pid_by_name
from client import Client
import re
from colorama import init, Fore, Back, Style
from logger import setup_logger
from git import Repo
import git

init()

_REMOTE_RESOURCE = 'pypeep_resource'
_VIRTUAL_ENV_NAME = 'virtual_decade'
_LOGGER = setup_logger('Main', color=Fore.BLUE)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--remote-path",
                        help="project path on remote client")
    parser.add_argument("--entry",
                        help="the entry python file of source code, or a executable file in the remote")
    parser.add_argument("--server-name", default='decade',
                        help="IDE server name (optional, default decade)")
    parser.add_argument("--hostname",
                        help="remote client hostname or docker container id")
    parser.add_argument("--ssh-user",
                        help="remote client ssh user, do not set if is docker container")
    parser.add_argument("--ssh-password",
                        help="remote client ssh password, do not set if is docker container")
    parser.add_argument("--ssh-port",
                        help="remote client ssh port (optional, default 22)", type=int, default=22)
    parser.add_argument("--local-path",
                        help="project path on local server, will download from remote if not exist")
    parser.add_argument("--venv",
                        help="specify virtualenv in local for package mapping, if not set, will use current python env")
    arguments = parser.parse_args()
    return arguments


# setup virtualenv
def setup_virtualenv(client, local_project_path, src_entry, remote_path):
    new_cmd = 'virtualenv {0}'.format(_VIRTUAL_ENV_NAME)
    client.execute(new_cmd)

    activate_cmd = 'source ./{0}/bin/activate'.format(_VIRTUAL_ENV_NAME)
    client.execute(activate_cmd)

    def _find_requirements_until_root(root_dir, sub_dir):
        assert root_dir in sub_dir

        if os.path.exists(os.path.join(sub_dir, 'requirements.txt')):
            return os.path.join(sub_dir.replace(root_dir, ''), 'requirements.txt')
        else:
            return _find_requirements_until_root(root_dir, os.path.dirname(sub_dir))

    # Install pydevd in remote
    client.execute('pip install pydevd')

    # Check the direct folder of the src entry file first, if no requirements.txt, then check the upper folder...
    src_path = os.path.join(local_project_path, *re.split(r'[/\\]*', src_entry))
    if os.path.exists(os.path.join(local_project_path, 'requirements.txt')):
        config_cmd = 'pip install -r {0}'.format(
            os.path.join(remote_path, _find_requirements_until_root(local_project_path, src_path)))
        client.execute(config_cmd)


def inject_sitecustomize(commands, client, local_ip, local_port):
    # hijack export keyword
    commands.append('source /tmp/{0}/hijack_export.sh'.format(_REMOTE_RESOURCE))
    # Declare the env variables in remote
    commands.append('export DECADE_IP={0};export DECADE_PORT={1}'.format(local_ip, local_port))


def edit_config_files(f, file_location, local_path, args_list):
    init_config = et.parse(file_location)
    root = init_config.getroot()
    for item in args_list:
        for ele in root.iter(item['tag']):
            for attrib_key in item['attrib'].keys():
                ele.set(attrib_key, item['attrib'][attrib_key])
    init_config.write(os.path.join(local_path, '.idea', f))


def git_check_version(local_project_path):
    if '.git' not in os.listdir(local_project_path):
        repo = Repo.init(local_project_path, bare=False)
    else:
        repo = Repo(local_project_path)

    commits_ahead = repo.iter_commits('origin/' + repo.active_branch.name + '..' + repo.active_branch.name)
    count_ahead = sum(1 for c in commits_ahead)
    if count_ahead:
        current_head = git.refs.head.HEAD(repo, path='HEAD')
        git.refs.head.HEAD.reset(current_head, commit='HEAD~' + str(count_ahead))

    origin = repo.remote()
    if repo.is_dirty():
        repo.git.stash('save')
        origin.fetch()
        origin.pull()


def config_IDE(args, remote_path, project_name, local_path, local_ip, local_port, ssh_port):
    local_idea_path = os.path.join(local_path, '.idea')

    if os.path.exists(local_idea_path):
        shutil.rmtree(local_idea_path)
    git_check_version(local_path)
    os.mkdir(local_idea_path)

    # if not os.path.exists(local_idea_path):
    #     os.mkdir(local_idea_path)
    # else:
    #     shutil.rmtree(local_idea_path)
    #     os.mkdir(local_idea_path)

    script_path = pkgutil.get_loader("decade").filename
    print script_path

    # other config files
    pycharm_config_dir = os.path.join(script_path, 'pycharm_config')
    raw_files = os.listdir(pycharm_config_dir)
    for f in raw_files:
        file_location = os.path.join(pycharm_config_dir, f)
        file_name = os.path.splitext(f)[0]
        if file_name == 'workspace' or file_name == 'webServer' or file_name == 'try':
            continue
        config_list = args[file_name]
        edit_config_files(f, file_location, local_path, config_list)

    # webServers.xml
    webservers_config = et.parse(os.path.join(pycharm_config_dir, 'webServers.xml'))
    webservers_root = webservers_config.getroot()
    for ele in webservers_root.iter('option'):
        if 'name' in ele.attrib.keys() and ele.get('name') == "port":
            ele.attrib['value'] = str(ssh_port)
    webservers_config.write(os.path.join(pycharm_config_dir, 'webServers.xml'))
    edit_config_files('webServers.xml', os.path.join(pycharm_config_dir, 'webServers.xml'), local_path,
                      args['webServers'])

    # workspace.xml
    workspace_config = et.parse(os.path.join(pycharm_config_dir, 'workspace.xml'))
    workspace_root = workspace_config.getroot()
    for ele in workspace_root.iter('option'):
        if 'name' in ele.attrib.keys() and ele.get('name') == "myItemId":
            ele.attrib['value'] = project_name
    for ele in workspace_root.iter('component'):
        if 'name' in ele.attrib.keys() and ele.get('name') == "RunManager":
            ele.attrib['selected'] = 'Python Remote Debug.debug1'
            debugger_list = ele.findall('configuration')
            for debugger in debugger_list:
                if 'type' in debugger.attrib.keys() and debugger.get('type') == "PyRemoteDebugConfigurationType":
                    debugger.set('name', 'debug1')
                    for option in debugger.iter('option'):
                        if 'name' in option.attrib.keys() and option.get('name') == 'PORT':
                            option.set('value', str(local_port))
                        if 'name' in option.attrib.keys() and option.get('name') == 'HOST':
                            option.set('value', local_ip)
                        if 'name' in option.attrib.keys() and option.get('name') == 'pathMappings':
                            for mapping in option.iter('mapping'):
                                mapping.set('local-root', '$PROJECT_DIR$')
                                mapping.set('remote-root', remote_path)
    workspace_config.write(os.path.join(local_path, '.idea', 'workspace.xml'))

    # iml
    shutil.copyfile(os.path.join(pycharm_config_dir, 'try.iml'),
                    os.path.join(local_path, '.idea', project_name + '.iml'))


def main():
    args = parse_args()
    remote_path = args.remote_path or os.environ.get('DECADE_REMOTE_PATH')
    assert remote_path
    server_name = args.server_name
    ssh_port = args.ssh_port
    local_path = args.local_path or os.environ.get('DECADE_LOCAL_PATH')
    assert local_path
    assert os.path.isdir(local_path), "local project path is not a directory."
    local_ip = get_host_ip()
    local_port = get_unoccupied_port()
    project_name = os.path.basename(remote_path)

    ide_config = {
        "deployment": [
            {'tag': 'component', 'attrib': {'serverName': server_name}},
            {'tag': 'paths', 'attrib': {'name': server_name}},
            {'tag': 'mapping', 'attrib': {'deploy': remote_path, 'local': '$PROJECT_DIR$' + remote_path}}
        ],
        "misc": [
        ],
        "remote-mappings": [
            {'tag': 'mapping', 'attrib': {'local-root': '$PROJECT_DIR$' + remote_path, 'remote-root': remote_path}},
        ],
        "webServers": [
        ],
    }

    client = Client(args.hostname, args.ssh_user, args.ssh_password, args.ssh_port)

    client.send_files(os.path.join(pkgutil.get_loader("decade").filename, _REMOTE_RESOURCE),
                      os.path.join('/tmp', _REMOTE_RESOURCE))

    # remote project is placed in the local project path. Modify this for consistency
    # local project path is empty
    local_project_path = os.path.join(local_path, project_name)

    if not os.path.exists(local_project_path):
        client.fetch_files(remote_path, local_project_path)

    config_IDE(ide_config, remote_path, project_name, local_project_path, local_ip, local_port, ssh_port)

    commands = []

    inject_sitecustomize(commands, client, local_ip, local_port)

    # setup_virtualenv(client, local_project_path, args.entry, remote_path)

    call(['open', '-a', 'PyCharm', local_project_path])

    _LOGGER.info('>> Please start the debug server in the PyCharm to continue <<')

    # use a loop to check if the debugger started(if port is occupied).
    while 1:
        port_open = False
        pid_list = get_pid_by_name('pycharm')
        for pid in pid_list:
            port_open = port_open or is_port_in_use(pid, local_port)
        if port_open:
            break
        _LOGGER.info('Still waiting...')
        time.sleep(10)
    _LOGGER.info('Detect the debugging port is open, ready to start')

    if args.entry.endswith('.py'):
        commands.append('python {0}'.format(args.entry))
    else:
        commands.append('source {0}'.format(args.entry))
    client.execute('\n'.join(commands))


if __name__ == "__main__":
    main()



import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ['--cov=pypeep']

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


install_requirements = [
    "paramiko",
    "docker",
    "colorama",
    "psutil",
]

test_requirements = [
    "pytest",
    "pytest-cov",
    "coveralls",
]

setup(
    name="pypeep",
    version="0.1.1",
    keywords=("PyCharm", "debug", "remote debug"),
    description="debug configurator",
    long_description="Configure Pycharm for remote debugging",
    license="MIT",

    url="https://github.com/cuyu/pypeep",
    author="cuyu",
    author_email="cuyu@splunk.com",

    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 2.7'
    ],
    platforms="any",
    install_requires=install_requirements,
    cmdclass={
        'test': PyTest,
    },
    extras_require={
        'test': test_requirements,
    },
    entry_points={
        'console_scripts': ['pypeep = pypeep.main:main']
    }
)
